@@ -19,6 +19,18 @@ class MySrvList;
 class MyHGC;
 
 
+MySQL_Connection *MySrvConnList::index(unsigned int _k) {
+	return (MySQL_Connection *)conns->index(_k);
+}
+
+MySQL_Connection * MySrvConnList::remove(int _k) {
+	return (MySQL_Connection *)conns->remove_index_fast(_k);
+}
+
+unsigned int MySrvConnList::conns_length() {
+	return conns->len;
+}
+
 MySrvConnList::MySrvConnList(MySrvC *_mysrvc) {
 	mysrvc=_mysrvc;
 	conns=new PtrArray();
@@ -30,7 +42,7 @@ void MySrvConnList::add(MySQL_Connection *c) {
 
 MySrvConnList::~MySrvConnList() {
 	mysrvc=NULL;
-	while (conns->len) {
+	while (conns_length()) {
 		MySQL_Connection *conn=(MySQL_Connection *)conns->remove_index_fast(0);
 		delete conn;
 	}
@@ -78,8 +90,9 @@ void MySrvList::remove(MySrvC *s) {
 	servers->remove_index_fast((unsigned int)i);
 }
 
+/*
 int MySrvConnList::find_idx(MySQL_Connection *c) {
-  for (unsigned int i=0; i<conns->len; i++) {
+  for (unsigned int i=0; i<conns_length(); i++) {
     MySQL_Connection *conn=(MySQL_Connection *)conns->index(i);
     if (conn==c) {
       return (unsigned int)i;
@@ -87,16 +100,19 @@ int MySrvConnList::find_idx(MySQL_Connection *c) {
   }
   return -1;
 }
+*/
 
+/*
 void MySrvConnList::remove(MySQL_Connection *c) {
 	int i=find_idx(c);
 	assert(i>=0);
 	conns->remove_index_fast((unsigned int)i);
 }
+*/
 
 void MySrvConnList::drop_all_connections() {
-	proxy_debug(PROXY_DEBUG_MYSQL_CONNPOOL, 7, "Dropping all connections (%lu total) on MySrvConnList %p for server %s:%d , hostgroup=%d , status=%d\n", conns->len, this, mysrvc->address, mysrvc->port, mysrvc->myhgc->hid, mysrvc->status);
-	while (conns->len) {
+	proxy_debug(PROXY_DEBUG_MYSQL_CONNPOOL, 7, "Dropping all connections (%lu total) on MySrvConnList %p for server %s:%d , hostgroup=%d , status=%d\n", conns_length(), this, mysrvc->address, mysrvc->port, mysrvc->myhgc->hid, mysrvc->status);
+	while (conns_length()) {
 		MySQL_Connection *conn=(MySQL_Connection *)conns->remove_index_fast(0);
 		delete conn;
 	}
@@ -140,6 +156,7 @@ void MySrvC::connect_error(int err_num) {
 		case 1044: // access denied
 		case 1045: // access denied
 		case 1049: //Unknown databas
+               case 1226: // do not increase counter to avoid Shunning if user has max_user_connections restriction
 			return;
 			break;
 		default:
@@ -263,7 +280,11 @@ MySQL_HostGroups_Manager::MySQL_HostGroups_Manager() {
 	status.autocommit_cnt_filtered=0;
 	status.commit_cnt_filtered=0;
 	status.rollback_cnt_filtered=0;
+#ifdef MHM_PTHREAD_MUTEX
+	pthread_mutex_init(&lock, NULL);
+#else
 	spinlock_rwlock_init(&rwlock);
+#endif
 	admindb=NULL;	// initialized only if needed
 	mydb=new SQLite3DB();
 	mydb->open((char *)"file:mem_mydb?mode=memory&cache=shared", SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE | SQLITE_OPEN_FULLMUTEX);
@@ -284,8 +305,12 @@ MySQL_HostGroups_Manager::~MySQL_HostGroups_Manager() {
 	if (admindb) {
 		delete admindb;
 	}
+#ifdef MHM_PTHREAD_MUTEX
+	pthread_mutex_destroy(&lock);
+#endif
 }
 
+/*
 void MySQL_HostGroups_Manager::rdlock() {
 	spin_wrlock(&rwlock);
 }
@@ -293,15 +318,23 @@ void MySQL_HostGroups_Manager::rdlock() {
 void MySQL_HostGroups_Manager::rdunlock() {
 	spin_wrunlock(&rwlock);
 }
-
+*/
 
 // wrlock() is only required during commit()
 void MySQL_HostGroups_Manager::wrlock() {
+#ifdef MHM_PTHREAD_MUTEX
+	pthread_mutex_lock(&lock);
+#else
 	spin_wrlock(&rwlock);
+#endif
 }
 
 void MySQL_HostGroups_Manager::wrunlock() {
+#ifdef MHM_PTHREAD_MUTEX
+	pthread_mutex_unlock(&lock);
+#else
 	spin_wrunlock(&rwlock);
+#endif
 }
 
 unsigned int MySQL_HostGroups_Manager::get_servers_table_version() {
@@ -427,8 +460,8 @@ bool MySQL_HostGroups_Manager::commit() {
 					proxy_debug(PROXY_DEBUG_MYSQL_CONNPOOL, 5, "Changing max_latency_ms for server %s:%d (%s:%d) from %d (%d) to %d\n" , mysrvc->address, mysrvc->port, r->fields[1], atoi(r->fields[2]), r->fields[9] , mysrvc->max_latency_us , atoi(r->fields[18]));
 					mysrvc->max_latency_us=1000*atoi(r->fields[18]);
 				}
-				if (atoi(r->fields[10])!=atoi(r->fields[19])) {
-					proxy_debug(PROXY_DEBUG_MYSQL_CONNPOOL, 5, "Changing comment for server %s:%d (%s:%d) from %d (%d) to %d\n" , mysrvc->address, mysrvc->port, r->fields[1], atoi(r->fields[2]), r->fields[10] , mysrvc->max_latency_us , atoi(r->fields[19]));
+				if (strcmp(r->fields[10],r->fields[19])) {
+					proxy_debug(PROXY_DEBUG_MYSQL_CONNPOOL, 5, "Changing comment for server %s:%d (%s:%d) from '%s' to '%s'\n" , mysrvc->address, mysrvc->port, r->fields[1], atoi(r->fields[2]), r->fields[10], r->fields[19]);
 					free(mysrvc->comment);
 					mysrvc->comment=strdup(r->fields[19]);
 				}
@@ -464,7 +497,7 @@ void MySQL_HostGroups_Manager::purge_mysql_servers_table() {
 		for (unsigned int j=0; j<myhgc->mysrvs->servers->len; j++) {
 			mysrvc=myhgc->mysrvs->idx(j);
 			if (mysrvc->status==MYSQL_SERVER_STATUS_OFFLINE_HARD) {
-				if (mysrvc->ConnectionsUsed->conns->len==0 && mysrvc->ConnectionsFree->conns->len==0) {
+				if (mysrvc->ConnectionsUsed->conns_length()==0 && mysrvc->ConnectionsFree->conns_length()==0) {
 					// no more connections for OFFLINE_HARD server, removing it
 					mysrvc=(MySrvC *)myhgc->mysrvs->servers->remove_index_fast(j);
 					delete mysrvc;
@@ -635,8 +668,13 @@ void MySQL_HostGroups_Manager::push_MyConn_to_pool(MySQL_Connection *c, bool _lo
 	}
 	if (mysrvc->status==MYSQL_SERVER_STATUS_ONLINE) {
 		if (c->async_state_machine==ASYNC_IDLE) {
-			c->optimize();
-			mysrvc->ConnectionsFree->add(c);
+			if (c->local_stmts->get_num_entries() > (unsigned int)mysql_thread___max_stmts_per_connection) {
+				proxy_debug(PROXY_DEBUG_MYSQL_CONNPOOL, 7, "Destroying MySQL_Connection %p, server %s:%d with status %d because has too many prepared statements\n", c, mysrvc->address, mysrvc->port, mysrvc->status);
+				delete c;
+			} else {
+				c->optimize();
+				mysrvc->ConnectionsFree->add(c);
+			}
 		} else {
 			proxy_debug(PROXY_DEBUG_MYSQL_CONNPOOL, 7, "Destroying MySQL_Connection %p, server %s:%d with status %d\n", c, mysrvc->address, mysrvc->port, mysrvc->status);
 			delete c;
@@ -674,10 +712,10 @@ MySrvC *MyHGC::get_random_MySrvC() {
 		for (j=0; j<l; j++) {
 			mysrvc=mysrvs->idx(j);
 			if (mysrvc->status==MYSQL_SERVER_STATUS_ONLINE) { // consider this server only if ONLINE
-				if (mysrvc->ConnectionsUsed->conns->len < mysrvc->max_connections) { // consider this server only if didn't reach max_connections
+				if (mysrvc->ConnectionsUsed->conns_length() < mysrvc->max_connections) { // consider this server only if didn't reach max_connections
 					if ( mysrvc->current_latency_us < ( mysrvc->max_latency_us ? mysrvc->max_latency_us : mysql_thread___default_max_latency_ms*1000 ) ) { // consider the host only if not too far
 						sum+=mysrvc->weight;
-						TotalUsedConn+=mysrvc->ConnectionsUsed->conns->len;
+						TotalUsedConn+=mysrvc->ConnectionsUsed->conns_length();
 					}
 				}
 			} else {
@@ -699,7 +737,7 @@ MySrvC *MyHGC::get_random_MySrvC() {
 							if (
 								(mysrvc->shunned_and_kill_all_connections==false) // it is safe to bring it back online
 								||
-								(mysrvc->shunned_and_kill_all_connections==true && mysrvc->ConnectionsUsed->conns->len==0 && mysrvc->ConnectionsFree->conns->len==0) // if shunned_and_kill_all_connections is set, ensure all connections are already dropped
+								(mysrvc->shunned_and_kill_all_connections==true && mysrvc->ConnectionsUsed->conns_length()==0 && mysrvc->ConnectionsFree->conns_length()==0) // if shunned_and_kill_all_connections is set, ensure all connections are already dropped
 							) {
 								mysrvc->status=MYSQL_SERVER_STATUS_ONLINE;
 								mysrvc->shunned_automatic=false;
@@ -709,7 +747,7 @@ MySrvC *MyHGC::get_random_MySrvC() {
 								// if a server is taken back online, consider it immediately
 								if ( mysrvc->current_latency_us < ( mysrvc->max_latency_us ? mysrvc->max_latency_us : mysql_thread___default_max_latency_ms*1000 ) ) { // consider the host only if not too far
 									sum+=mysrvc->weight;
-									TotalUsedConn+=mysrvc->ConnectionsUsed->conns->len;
+									TotalUsedConn+=mysrvc->ConnectionsUsed->conns_length();
 								}
 							}
 						}
@@ -738,7 +776,7 @@ MySrvC *MyHGC::get_random_MySrvC() {
 						// if a server is taken back online, consider it immediately
 						if ( mysrvc->current_latency_us < ( mysrvc->max_latency_us ? mysrvc->max_latency_us : mysql_thread___default_max_latency_ms*1000 ) ) { // consider the host only if not too far
 							sum+=mysrvc->weight;
-							TotalUsedConn+=mysrvc->ConnectionsUsed->conns->len;
+							TotalUsedConn+=mysrvc->ConnectionsUsed->conns_length();
 						}
 					}
 				}
@@ -756,7 +794,7 @@ MySrvC *MyHGC::get_random_MySrvC() {
 		for (j=0; j<l; j++) {
 			mysrvc=mysrvs->idx(j);
 			if (mysrvc->status==MYSQL_SERVER_STATUS_ONLINE) { // consider this server only if ONLINE
-				unsigned int len=mysrvc->ConnectionsUsed->conns->len;
+				unsigned int len=mysrvc->ConnectionsUsed->conns_length();
 				if (len < mysrvc->max_connections) { // consider this server only if didn't reach max_connections
 					if ( mysrvc->current_latency_us < ( mysrvc->max_latency_us ? mysrvc->max_latency_us : mysql_thread___default_max_latency_ms*1000 ) ) { // consider the host only if not too far
 						if ((len * sum) <= (TotalUsedConn * mysrvc->weight * 1.5 + 1)) {
@@ -773,14 +811,20 @@ MySrvC *MyHGC::get_random_MySrvC() {
 			return NULL; // if we reach here, we couldn't find any target
 		}
 
-		unsigned int k=rand()%New_sum;
+		//unsigned int k=rand()%New_sum;
+		unsigned int k;
+		if (New_sum > 32768) {
+			k=rand()%New_sum;
+		} else {
+			k=fastrand()%New_sum;
+		}
   	k++;
 		New_sum=0;
 
 		for (j=0; j<l; j++) {
 			mysrvc=mysrvs->idx(j);
 			if (mysrvc->status==MYSQL_SERVER_STATUS_ONLINE) { // consider this server only if ONLINE
-				unsigned int len=mysrvc->ConnectionsUsed->conns->len;
+				unsigned int len=mysrvc->ConnectionsUsed->conns_length();
 				if (len < mysrvc->max_connections) { // consider this server only if didn't reach max_connections
 					if ( mysrvc->current_latency_us < ( mysrvc->max_latency_us ? mysrvc->max_latency_us : mysql_thread___default_max_latency_ms*1000 ) ) { // consider the host only if not too far
 						if ((len * sum) <= (TotalUsedConn * mysrvc->weight * 1.5 + 1)) {
@@ -814,9 +858,14 @@ MySrvC * MySrvList::idx(unsigned int i) { return (MySrvC *)servers->index(i); }
 MySQL_Connection * MySrvConnList::get_random_MyConn() {
 	MySQL_Connection * conn=NULL;
 	unsigned int i;
-	unsigned int l=conns->len;
+	unsigned int l=conns_length();
 	if (l) {
-		i=rand()%l;
+		//i=rand()%l;
+		if (l>32768) {
+			i=rand()%l;
+		} else {
+			i=fastrand()%l;
+		}
 		conn=(MySQL_Connection *)conns->remove_index_fast(i);
 		proxy_debug(PROXY_DEBUG_MYSQL_CONNPOOL, 7, "Returning MySQL Connection %p, server %s:%d\n", conn, conn->parent->address, conn->parent->port);
 		return conn;
@@ -896,7 +945,11 @@ void MySQL_HostGroups_Manager::replication_lag_action(int _hid, char *address, u
 					}
 				} else {
 					if (mysrvc->status==MYSQL_SERVER_STATUS_SHUNNED_REPLICATION_LAG) {
-						if (current_replication_lag>=0 && ((unsigned int)current_replication_lag <= mysrvc->max_replication_lag)) {
+						if (
+							(current_replication_lag>=0 && ((unsigned int)current_replication_lag <= mysrvc->max_replication_lag))
+							||
+							(current_replication_lag==-2) // see issue 959
+						) {
 							mysrvc->status=MYSQL_SERVER_STATUS_ONLINE;
 						}
 					}
@@ -923,15 +976,16 @@ void MySQL_HostGroups_Manager::drop_all_idle_connections() {
 			}
 
 			// Drop idle connections if beyond max_connection
-			while (mysrvc->ConnectionsFree->conns->len && mysrvc->ConnectionsUsed->conns->len+mysrvc->ConnectionsFree->conns->len > mysrvc->max_connections) {
-				MySQL_Connection *conn=(MySQL_Connection *)mysrvc->ConnectionsFree->conns->remove_index_fast(0);
+			while (mysrvc->ConnectionsFree->conns_length() && mysrvc->ConnectionsUsed->conns_length()+mysrvc->ConnectionsFree->conns_length() > mysrvc->max_connections) {
+				MySQL_Connection *conn=mysrvc->ConnectionsFree->remove(0);
 				delete conn;
 				//__sync_fetch_and_sub(&status.server_connections_connected, 1);
 			}
 
-			PtrArray *pa=mysrvc->ConnectionsFree->conns;
-			while (pa->len > mysql_thread___free_connections_pct*mysrvc->max_connections/100) {
-				MySQL_Connection *mc=(MySQL_Connection *)pa->remove_index_fast(0);
+			//PtrArray *pa=mysrvc->ConnectionsFree->conns;
+			MySrvConnList *mscl=mysrvc->ConnectionsFree;
+			while (mscl->conns_length() > mysql_thread___free_connections_pct*mysrvc->max_connections/100) {
+				MySQL_Connection *mc=mscl->remove(0);
 				delete mc;
 				//__sync_fetch_and_sub(&status.server_connections_connected, 1);
 			}
@@ -940,10 +994,10 @@ void MySQL_HostGroups_Manager::drop_all_idle_connections() {
 			if (mysql_thread___connection_max_age_ms) {
 				unsigned long long curtime=monotonic_time();
 				int i=0;
-				for (i=0; i<(int)pa->len ; i++) {
-					MySQL_Connection *mc=(MySQL_Connection *)pa->index(i);
+				for (i=0; i<(int)mscl->conns_length() ; i++) {
+					MySQL_Connection *mc=mscl->index(i);
 					if (curtime > mc->creation_time + mysql_thread___connection_max_age_ms * 1000) {
-						mc=(MySQL_Connection *)pa->remove_index_fast(0);
+						mc=mscl->remove(0);
 						delete mc;
 						i--;
 					}
@@ -978,12 +1032,14 @@ int MySQL_HostGroups_Manager::get_multiple_idle_connections(int _hid, unsigned l
 		if (_hid >= 0 && _hid!=(int)myhgc->hid) continue;
 		for (j=0; j<(int)myhgc->mysrvs->cnt(); j++) {
 			MySrvC *mysrvc=(MySrvC *)myhgc->mysrvs->servers->index(j);
-			PtrArray *pa=mysrvc->ConnectionsFree->conns;
-			for (k=0; k<(int)pa->len; k++) {
-				MySQL_Connection *mc=(MySQL_Connection *)pa->index(k);
+			//PtrArray *pa=mysrvc->ConnectionsFree->conns;
+			MySrvConnList *mscl=mysrvc->ConnectionsFree;
+			for (k=0; k<(int)mscl->conns_length(); k++) {
+				MySQL_Connection *mc=mscl->index(k);
 				// If the connection is idle ...
 				if (mc->last_time_used < _max_last_time_used) {
-					mc=(MySQL_Connection *)pa->remove_index_fast(k);
+					//mc=(MySQL_Connection *)pa->remove_index_fast(k);
+					mc=mscl->remove(k);
 					mysrvc->ConnectionsUsed->add(mc);
 					k--;
 					conn_list[num_conn_current]=mc;
@@ -1032,8 +1088,9 @@ SQLite3_result * MySQL_HostGroups_Manager::SQL3_Connection_Pool() {
 				mysrvc->ConnectionsFree->drop_all_connections();
 			}
 			// drop idle connections if beyond max_connection
-			while (mysrvc->ConnectionsFree->conns->len && mysrvc->ConnectionsUsed->conns->len+mysrvc->ConnectionsFree->conns->len > mysrvc->max_connections) {
-				MySQL_Connection *conn=(MySQL_Connection *)mysrvc->ConnectionsFree->conns->remove_index_fast(0);
+			while (mysrvc->ConnectionsFree->conns_length() && mysrvc->ConnectionsUsed->conns_length()+mysrvc->ConnectionsFree->conns_length() > mysrvc->max_connections) {
+				//MySQL_Connection *conn=(MySQL_Connection *)mysrvc->ConnectionsFree->conns->remove_index_fast(0);
+				MySQL_Connection *conn=mysrvc->ConnectionsFree->remove(0);
 				delete conn;
 				//__sync_fetch_and_sub(&status.server_connections_connected, 1);
 			}
@@ -1064,9 +1121,9 @@ SQLite3_result * MySQL_HostGroups_Manager::SQL3_Connection_Pool() {
 					assert(0);
 					break;
 			}
-			sprintf(buf,"%u", mysrvc->ConnectionsUsed->conns->len);
+			sprintf(buf,"%u", mysrvc->ConnectionsUsed->conns_length());
 			pta[4]=strdup(buf);
-			sprintf(buf,"%u", mysrvc->ConnectionsFree->conns->len);
+			sprintf(buf,"%u", mysrvc->ConnectionsFree->conns_length());
 			pta[5]=strdup(buf);
 			sprintf(buf,"%u", mysrvc->connect_OK);
 			pta[6]=strdup(buf);
@@ -1127,6 +1184,8 @@ void MySQL_HostGroups_Manager::read_only_action(char *hostname, int port, int re
 //		num_rows=resultset->rows_count;
 //	}
 
+	delete resultset;
+	resultset=NULL;
 	if (admindb==NULL) { // we initialize admindb only if needed
 		admindb=new SQLite3DB();
 		admindb->open((char *)"file:mem_admindb?mode=memory&cache=shared", SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE | SQLITE_OPEN_FULLMUTEX);	
@@ -1281,9 +1340,10 @@ unsigned long long MySQL_HostGroups_Manager::Get_Memory_Stats() {
 			for (j=0; j<l; j++) {
 				intsize+=sizeof(MySrvC);
 				mysrvc=myhgc->mysrvs->idx(j);
-				intsize+=((mysrvc->ConnectionsUsed->conns->size)*sizeof(MySQL_Connection *));
-				for (k=0; k<mysrvc->ConnectionsFree->conns->len; k++) {
-					MySQL_Connection *myconn=(MySQL_Connection *)mysrvc->ConnectionsFree->conns->index(k);
+				intsize+=((mysrvc->ConnectionsUsed->conns_length())*sizeof(MySQL_Connection *));
+				for (k=0; k<mysrvc->ConnectionsFree->conns_length(); k++) {
+					//MySQL_Connection *myconn=(MySQL_Connection *)mysrvc->ConnectionsFree->conns->index(k);
+					MySQL_Connection *myconn=mysrvc->ConnectionsFree->index(k);
 					intsize+=sizeof(MySQL_Connection)+sizeof(MYSQL);
 					intsize+=myconn->mysql->net.max_packet;
 					intsize+=(4096*15); // ASYNC_CONTEXT_DEFAULT_STACK_SIZE
@@ -1291,7 +1351,7 @@ unsigned long long MySQL_HostGroups_Manager::Get_Memory_Stats() {
 						intsize+=myconn->MyRS->current_size();
 					}
 				}
-				intsize+=((mysrvc->ConnectionsUsed->conns->size)*sizeof(MySQL_Connection *));
+				intsize+=((mysrvc->ConnectionsUsed->conns_length())*sizeof(MySQL_Connection *));
 			}
 		}
 	}
