@@ -153,20 +153,15 @@ MySQL_Monitor_Connection_Pool::~MySQL_Monitor_Connection_Pool() {
 void MySQL_Monitor_Connection_Pool::purge_idle_connections() {
 	unsigned long long now=monotonic_time();
 	pthread_mutex_lock(&mutex);
-	std::map<char *, std::list<MYSQL *>*>::iterator it;
 	//fprintf(stderr,"conn pool size: %d\n",my_connections.size());
-	unsigned int totconn;
-	totconn=0;
-	for(it = my_connections.begin(); it != my_connections.end(); it++) {
-		std::list<MYSQL *> *lst=it->second;
-		totconn+=lst->size();
-	}
+	unsigned int totconn = 0;
+	for(auto& it : my_connections)
+		totconn+=it.second->size();
 	//fprintf(stderr,"tot conn in pool: %d\n",totconn);
-	for(it = my_connections.begin(); it != my_connections.end(); it++) {
+	for(auto it = my_connections.begin(); it != my_connections.end();) {
 		std::list<MYSQL *> *lst=it->second;
 		if (!lst->empty()) {
-			std::list<MYSQL *>::const_iterator it3;
-			for(it3 = lst->begin(); it3 != lst->end(); it3++) {
+			for(auto it3 = lst->begin(); it3 != lst->end();) {
 				//it3=lst->begin();
 				MYSQL *my=*it3;
 				unsigned long long then=0;
@@ -177,11 +172,15 @@ void MySQL_Monitor_Connection_Pool::purge_idle_connections() {
 					WorkItem *item;
 					item=new WorkItem(mmsd,NULL);
 					GloMyMon->queue.add(item);
-					lst->remove(*it3);
-				}
+					it3 = lst->erase(it3);
+				} else
+					++it3;
 			}
+			++it;
 		} else {
-			my_connections.erase(it);
+			free(it->first);
+			free(it->second);
+			it = my_connections.erase(it);
 		}
 	}
 	pthread_mutex_unlock(&mutex);
@@ -262,6 +261,34 @@ MySQL_Monitor_State_Data::~MySQL_Monitor_State_Data() {
 	}
 }
 
+void * monitor_connect_pthread(void *arg) {
+	bool cache=false;
+	mallctl("thread.tcache.enabled", NULL, NULL, &cache, sizeof(bool));
+	GloMyMon->monitor_connect();
+	return NULL;
+}
+
+void * monitor_ping_pthread(void *arg) {
+	bool cache=false;
+	mallctl("thread.tcache.enabled", NULL, NULL, &cache, sizeof(bool));
+	GloMyMon->monitor_ping();
+	return NULL;
+}
+
+void * monitor_read_only_pthread(void *arg) {
+	bool cache=false;
+	mallctl("thread.tcache.enabled", NULL, NULL, &cache, sizeof(bool));
+	GloMyMon->monitor_read_only();
+	return NULL;
+}
+
+void * monitor_replication_lag_pthread(void *arg) {
+	bool cache=false;
+	mallctl("thread.tcache.enabled", NULL, NULL, &cache, sizeof(bool));
+	GloMyMon->monitor_replication_lag();
+	return NULL;
+}
+
 MySQL_Monitor::MySQL_Monitor() {
 
 	GloMyMon = this;
@@ -298,6 +325,9 @@ MySQL_Monitor::MySQL_Monitor() {
 			num_threads=GloMTH->num_threads*2;
 		}
 	}
+	if (num_threads>16) {
+		num_threads=16;	// limit to 16
+	}
 };
 
 MySQL_Monitor::~MySQL_Monitor() {
@@ -576,9 +606,9 @@ void * monitor_read_only_thread(void *arg) {
 	while (mmsd->async_exit_status) {
 		mmsd->async_exit_status=wait_for_mysql(mmsd->mysql, mmsd->async_exit_status);
 		unsigned long long now=monotonic_time();
-		if (now > mmsd->t1 + mysql_thread___monitor_ping_timeout * 1000) {
+		if (now > mmsd->t1 + mysql_thread___monitor_read_only_timeout * 1000) {
 			mmsd->mysql_error_msg=strdup("timeout check");
-			proxy_error("Timeout on read_only check for %s:%d. If the server is overload, increase mysql-monitor_read_only_timeout. Assuming read_only=1\n", mmsd->hostname, mmsd->port);
+			proxy_error("Timeout on read_only check for %s:%d after %lldms. If the server is overload, increase mysql-monitor_read_only_timeout. Assuming read_only=1\n", mmsd->hostname, mmsd->port, (now-mmsd->t1)/1000);
 			goto __exit_monitor_read_only_thread;
 		}
 		if (GloMyMon->shutdown==true) {
@@ -592,9 +622,9 @@ void * monitor_read_only_thread(void *arg) {
 	while (mmsd->async_exit_status) {
 		mmsd->async_exit_status=wait_for_mysql(mmsd->mysql, mmsd->async_exit_status);
 		unsigned long long now=monotonic_time();
-		if (now > mmsd->t1 + mysql_thread___monitor_ping_timeout * 1000) {
+		if (now > mmsd->t1 + mysql_thread___monitor_read_only_timeout * 1000) {
 			mmsd->mysql_error_msg=strdup("timeout check");
-			proxy_error("Timeout on read_only check for %s:%d. If the server is overload, increase mysql-monitor_read_only_timeout. Assuming read_only=1\n", mmsd->hostname, mmsd->port);
+			proxy_error("Timeout on read_only check for %s:%d after %lldms. If the server is overload, increase mysql-monitor_read_only_timeout. Assuming read_only=1\n", mmsd->hostname, mmsd->port, (now-mmsd->t1)/1000);
 			goto __exit_monitor_read_only_thread;
 		}
 		if (GloMyMon->shutdown==true) {
@@ -876,7 +906,6 @@ void * MySQL_Monitor::monitor_connect() {
 	unsigned long long t1;
 	unsigned long long t2;
 	unsigned long long next_loop_at=0;
-	unsigned long long start_time;
 	while (GloMyMon->shutdown==false && mysql_thread___monitor_enabled==true) {
 
 		char *error=NULL;
@@ -901,8 +930,6 @@ void * MySQL_Monitor::monitor_connect() {
 		}
 		next_loop_at=t1+1000*mysql_thread___monitor_connect_interval;
 
-		start_time=monotonic_time();
-
 		proxy_debug(PROXY_DEBUG_ADMIN, 4, "%s\n", query);
 		admindb->execute_statement(query, &error , &cols , &affected_rows , &resultset);
 		if (error) {
@@ -942,7 +969,8 @@ __end_monitor_connect_loop:
 				if (mysql_thread___monitor_ping_interval < 3600000)
 					mysql_thread___monitor_history = mysql_thread___monitor_ping_interval * (mysql_thread___monitor_ping_max_failures + 1 );
 			}
-			rc=sqlite3_bind_int64(statement, 1, start_time-mysql_thread___monitor_history*1000); assert(rc==SQLITE_OK);
+			unsigned long long time_now=realtime_time();
+			rc=sqlite3_bind_int64(statement, 1, time_now-(unsigned long long)mysql_thread___monitor_history*1000); assert(rc==SQLITE_OK);
 			SAFE_SQLITE3_STEP(statement);
 			rc=sqlite3_clear_bindings(statement); assert(rc==SQLITE_OK);
 			rc=sqlite3_reset(statement); assert(rc==SQLITE_OK);
@@ -986,7 +1014,6 @@ void * MySQL_Monitor::monitor_ping() {
 
 	unsigned long long t1;
 	unsigned long long t2;
-	unsigned long long start_time;
 	unsigned long long next_loop_at=0;
 
 	while (GloMyMon->shutdown==false && mysql_thread___monitor_enabled==true) {
@@ -1012,8 +1039,6 @@ void * MySQL_Monitor::monitor_ping() {
 		}
 		next_loop_at=t1+1000*mysql_thread___monitor_ping_interval;
 
-		start_time=monotonic_time();
-
 		proxy_debug(PROXY_DEBUG_ADMIN, 4, "%s\n", query);
 		admindb->execute_statement(query, &error , &cols , &affected_rows , &resultset);
 		if (error) {
@@ -1052,7 +1077,8 @@ __end_monitor_ping_loop:
 				if (mysql_thread___monitor_ping_interval < 3600000)
 					mysql_thread___monitor_history = mysql_thread___monitor_ping_interval * (mysql_thread___monitor_ping_max_failures + 1 );
 			}
-			rc=sqlite3_bind_int64(statement, 1, start_time-mysql_thread___monitor_history*1000); assert(rc==SQLITE_OK);
+			unsigned long long time_now=realtime_time();
+			rc=sqlite3_bind_int64(statement, 1, time_now-(unsigned long long)mysql_thread___monitor_history*1000); assert(rc==SQLITE_OK);
 			SAFE_SQLITE3_STEP(statement);
 			rc=sqlite3_clear_bindings(statement); assert(rc==SQLITE_OK);
 			rc=sqlite3_reset(statement); assert(rc==SQLITE_OK);
@@ -1165,6 +1191,13 @@ __end_monitor_ping_loop:
 				}
 				free(buff);
 			}
+			while (i) { // now free all the addresses/ports
+				i--;
+				free(addresses[i]);
+				free(ports[i]);
+			}
+			free(addresses);
+			free(ports);
 		}
 
 __sleep_monitor_ping_loop:
@@ -1201,7 +1234,6 @@ void * MySQL_Monitor::monitor_read_only() {
 
 	unsigned long long t1;
 	unsigned long long t2;
-	unsigned long long start_time;
 	unsigned long long next_loop_at=0;
 
 	while (GloMyMon->shutdown==false && mysql_thread___monitor_enabled==true) {
@@ -1225,7 +1257,6 @@ void * MySQL_Monitor::monitor_read_only() {
 		if (t1 < next_loop_at) {
 			goto __sleep_monitor_read_only;
 		}
-		start_time=monotonic_time();
 		next_loop_at=t1+1000*mysql_thread___monitor_read_only_interval;
 		proxy_debug(PROXY_DEBUG_ADMIN, 4, "%s\n", query);
 //		admindb->execute_statement(query, &error , &cols , &affected_rows , &resultset);
@@ -1271,7 +1302,8 @@ __end_monitor_read_only_loop:
 				if (mysql_thread___monitor_ping_interval < 3600000)
 					mysql_thread___monitor_history = mysql_thread___monitor_ping_interval * (mysql_thread___monitor_ping_max_failures + 1 );
 			}
-			rc=sqlite3_bind_int64(statement, 1, start_time-mysql_thread___monitor_history*1000); assert(rc==SQLITE_OK);
+			unsigned long long time_now=realtime_time();
+			rc=sqlite3_bind_int64(statement, 1, time_now-(unsigned long long)mysql_thread___monitor_history*1000); assert(rc==SQLITE_OK);
 			SAFE_SQLITE3_STEP(statement);
 			rc=sqlite3_clear_bindings(statement); assert(rc==SQLITE_OK);
 			rc=sqlite3_reset(statement); assert(rc==SQLITE_OK);
@@ -1318,7 +1350,6 @@ void * MySQL_Monitor::monitor_replication_lag() {
 
 	unsigned long long t1;
 	unsigned long long t2;
-	unsigned long long start_time;
 	unsigned long long next_loop_at=0;
 
 	while (GloMyMon->shutdown==false && mysql_thread___monitor_enabled==true) {
@@ -1343,7 +1374,6 @@ void * MySQL_Monitor::monitor_replication_lag() {
 			goto __sleep_monitor_replication_lag;
 		}
 		next_loop_at=t1+1000*mysql_thread___monitor_replication_lag_interval;
-		start_time=t1;
 
 		proxy_debug(PROXY_DEBUG_ADMIN, 4, "%s\n", query);
 //		admindb->execute_statement(query, &error , &cols , &affected_rows , &resultset);
@@ -1386,7 +1416,8 @@ __end_monitor_replication_lag_loop:
 				if (mysql_thread___monitor_ping_interval < 3600000)
 					mysql_thread___monitor_history = mysql_thread___monitor_ping_interval * (mysql_thread___monitor_ping_max_failures + 1 );
 			}
-			rc=sqlite3_bind_int64(statement, 1, start_time-mysql_thread___monitor_history*1000); assert(rc==SQLITE_OK);
+			unsigned long long time_now=realtime_time();
+			rc=sqlite3_bind_int64(statement, 1, time_now-(unsigned long long)mysql_thread___monitor_history*1000); assert(rc==SQLITE_OK);
 			SAFE_SQLITE3_STEP(statement);
 			rc=sqlite3_clear_bindings(statement); assert(rc==SQLITE_OK);
 			rc=sqlite3_reset(statement); assert(rc==SQLITE_OK);
@@ -1441,12 +1472,19 @@ __monitor_run:
 	ConsumerThread **threads= (ConsumerThread **)malloc(sizeof(ConsumerThread *)*num_threads);
 	for (unsigned int i=0;i<num_threads; i++) {
 		threads[i] = new ConsumerThread(queue, 0);
-		threads[i]->start();
+		threads[i]->start(false);
 	}
-	std::thread * monitor_connect_thread = new std::thread(&MySQL_Monitor::monitor_connect,this);
-	std::thread * monitor_ping_thread = new std::thread(&MySQL_Monitor::monitor_ping,this);
-	std::thread * monitor_read_only_thread = new std::thread(&MySQL_Monitor::monitor_read_only,this);
-	std::thread * monitor_replication_lag_thread = new std::thread(&MySQL_Monitor::monitor_replication_lag,this);
+	pthread_attr_t attr;
+	pthread_attr_init(&attr);
+	pthread_attr_setstacksize (&attr, 64*1024);
+	pthread_t monitor_connect_thread;
+	pthread_create(&monitor_connect_thread, &attr, &monitor_connect_pthread,NULL);
+	pthread_t monitor_ping_thread;
+	pthread_create(&monitor_ping_thread, &attr, &monitor_ping_pthread,NULL);
+	pthread_t monitor_read_only_thread;
+	pthread_create(&monitor_read_only_thread, &attr, &monitor_read_only_pthread,NULL);
+	pthread_t monitor_replication_lag_thread;
+	pthread_create(&monitor_replication_lag_thread, &attr, &monitor_replication_lag_pthread,NULL);
 	while (shutdown==false && mysql_thread___monitor_enabled==true) {
 		unsigned int glover;
 		if (GloMTH) {
@@ -1485,10 +1523,10 @@ __monitor_run:
 		threads[i]->join();
 	}
 	free(threads);
-	monitor_connect_thread->join();
-	monitor_ping_thread->join();
-	monitor_read_only_thread->join();
-	monitor_replication_lag_thread->join();
+	pthread_join(monitor_connect_thread,NULL);
+	pthread_join(monitor_ping_thread,NULL);
+	pthread_join(monitor_read_only_thread,NULL);
+	pthread_join(monitor_replication_lag_thread,NULL);
 	while (shutdown==false) {
 		unsigned int glover;
 		if (GloMTH) {
