@@ -2,7 +2,7 @@
 #include "proxysql.h"
 #include "cpp.h"
 #include <string>
-
+#include <sys/utsname.h>
 
 
 static void term_handler(int sig) {
@@ -58,6 +58,9 @@ ProxySQL_GlobalVariables::ProxySQL_GlobalVariables() {
 	global.nostart=false;
 	global.foreground=false;
 	global.monitor=true;
+#ifdef SO_REUSEPORT
+		global.reuseport=false;
+#endif /* SO_REUSEPORT */
 //	global.use_proxysql_mem=false;
 	pthread_mutex_init(&global.start_mutex,NULL);
 #ifdef DEBUG
@@ -78,6 +81,9 @@ ProxySQL_GlobalVariables::ProxySQL_GlobalVariables() {
 	opt->add((const char *)"",0,0,0,(const char *)"Starts only the admin service",(const char *)"-n",(const char *)"--no-start");
 	opt->add((const char *)"",0,0,0,(const char *)"Do not start Monitor Module",(const char *)"-M",(const char *)"--no-monitor");
 	opt->add((const char *)"",0,0,0,(const char *)"Run in foreground",(const char *)"-f",(const char *)"--foreground");
+#ifdef SO_REUSEPORT
+	opt->add((const char *)"",0,0,0,(const char *)"Use SO_REUSEPORT",(const char *)"-r",(const char *)"--reuseport");
+#endif /* SO_REUSEPORT */
 	opt->add((const char *)"",0,0,0,(const char *)"Do not restart ProxySQL if crashes",(const char *)"-e",(const char *)"--exit-on-error");
 	opt->add((const char *)"~/proxysql.cnf",0,1,0,(const char *)"Configuraton file",(const char *)"-c",(const char *)"--config");
 	//opt->add((const char *)"",0,0,0,(const char *)"Enable custom memory allocator",(const char *)"-m",(const char *)"--custom-memory");
@@ -189,6 +195,33 @@ void ProxySQL_GlobalVariables::process_opts_post() {
 		global.monitor=false;
 	}
 
+#ifdef SO_REUSEPORT
+	{
+		struct utsname unameData;
+		int rc;
+		rc=uname(&unameData);
+		if (rc==0) {
+			if (strcmp(unameData.sysname,"Linux")==0) {
+				int major=0, minor=0, revision=0;
+				sscanf(unameData.release, "%d.%d.%d", &major, &minor, &revision);
+				//fprintf(stderr,"%d %d %d\n",major,minor,revision);
+				if (
+					(major > 3)
+					||
+					(major == 3 && minor >= 9)
+				) {
+					global.reuseport=true;
+				}
+			}
+		}
+	}
+#endif /* SO_REUSEPORT */
+#ifdef SO_REUSEPORT
+	if (opt->isSet("-r")) {
+		global.reuseport=true;
+	}
+#endif /* SO_REUSEPORT */
+
 	if (opt->isSet("-S")) {
 		std::string admin_socket;
 		opt->get("-S")->getString(admin_socket);
