@@ -123,7 +123,7 @@ bool MySQL_Connection_userinfo::set_schemaname(char *_new, int l) {
 	if (schemaname) {
 		_l=strlen(schemaname); // bug fix for #609
 	}
-	if ((schemaname==NULL) || (strncmp(_new,schemaname, (l > _l ? l : _l) ))) {
+	if ((schemaname==NULL) || (l != _l) || (strncmp(_new,schemaname, l ))) {
 		if (schemaname) {
 			free(schemaname);
 			schemaname=NULL;
@@ -164,18 +164,28 @@ MySQL_Connection::MySQL_Connection() {
 	status_flags=0;
 	options.compression_min_length=0;
 	options.server_version=NULL;
+	options.last_set_autocommit=-1;	// -1 = never set
 	options.autocommit=true;
 	options.init_connect=NULL;
 	options.init_connect_sent=false;
+	options.sql_log_bin=1;	// default #818
+	options.sql_mode=NULL;	// #509
+	options.sql_mode_int=0;	// #509
+	options.time_zone=NULL;	// #819
+	options.time_zone_int=0;	// #819
 	compression_pkt_id=0;
 	mysql_result=NULL;
 	query.ptr=NULL;
 	query.length=0;
+	query.stmt=NULL;
+	query.stmt_meta=NULL;
+	query.stmt_result=NULL;
 	largest_query_length=0;
 	MyRS=NULL;
 	creation_time=0;
 	processing_multi_statement=false;
 	proxy_debug(PROXY_DEBUG_MYSQL_CONNPOOL, 4, "Creating new MySQL_Connection %p\n", this);
+	local_stmts=new MySQL_STMTs_local(false); // false by default, it is a backend
 };
 
 MySQL_Connection::~MySQL_Connection() {
@@ -206,6 +216,22 @@ MySQL_Connection::~MySQL_Connection() {
 	if (MyRS) {
 		delete MyRS;
 	}
+	if (local_stmts) {
+		delete local_stmts;
+	}
+	if (query.stmt) {
+		// we don't run mysql_stmt_close() : should be already destroyed
+//		mysql_stmt_close(query.stmt);
+		query.stmt=NULL;
+	}
+	if (options.sql_mode) {
+		free(options.sql_mode);
+		options.sql_mode=NULL;
+	}
+	if (options.time_zone) {
+		free(options.time_zone);
+		options.time_zone=NULL;
+	}
 };
 
 bool MySQL_Connection::set_autocommit(bool _ac) {
@@ -250,6 +276,14 @@ void MySQL_Connection::set_status_get_lock(bool v) {
 	}
 }
 
+void MySQL_Connection::set_status_found_rows(bool v) {
+	if (v) {
+		status_flags |= STATUS_MYSQL_CONNECTION_FOUND_ROWS;
+	} else {
+		status_flags &= ~STATUS_MYSQL_CONNECTION_FOUND_ROWS;
+	}
+}
+
 void MySQL_Connection::set_status_lock_tables(bool v) {
 	if (v) {
 		status_flags |= STATUS_MYSQL_CONNECTION_LOCK_TABLES;
@@ -282,6 +316,17 @@ void MySQL_Connection::set_status_prepared_statement(bool v) {
 	}
 }
 
+// pay attention here. set_status_sql_log_bin0 sets it sql_log_bin is ZERO
+// sql_log_bin=0 => true
+// sql_log_bin=1 => false
+void MySQL_Connection::set_status_sql_log_bin0(bool v) {
+	if (v) {
+		status_flags |= STATUS_MYSQL_CONNECTION_SQL_LOG_BIN0;
+	} else {
+		status_flags &= ~STATUS_MYSQL_CONNECTION_SQL_LOG_BIN0;
+	}
+}
+
 bool MySQL_Connection::get_status_transaction() {
 	return status_flags & STATUS_MYSQL_CONNECTION_TRANSACTION;
 }
@@ -298,6 +343,10 @@ bool MySQL_Connection::get_status_get_lock() {
 	return status_flags & STATUS_MYSQL_CONNECTION_GET_LOCK;
 }
 
+bool MySQL_Connection::get_status_found_rows() {
+	return status_flags & STATUS_MYSQL_CONNECTION_FOUND_ROWS;
+}
+
 bool MySQL_Connection::get_status_lock_tables() {
 	return status_flags & STATUS_MYSQL_CONNECTION_LOCK_TABLES;
 }
@@ -310,6 +359,10 @@ bool MySQL_Connection::get_status_prepared_statement() {
 	return status_flags & STATUS_MYSQL_CONNECTION_PREPARED_STATEMENT;
 }
 
+bool MySQL_Connection::get_status_sql_log_bin0() {
+	return status_flags & STATUS_MYSQL_CONNECTION_SQL_LOG_BIN0;
+}
+
 // non blocking API
 void MySQL_Connection::connect_start() {
 	PROXY_TRACE();
@@ -431,6 +484,10 @@ void MySQL_Connection::set_query(char *stmt, unsigned long length) {
 	if (length > largest_query_length) {
 		largest_query_length=length;
 	}
+	if (query.stmt) {
+		//mysql_stmt_close(query.stmt);
+		query.stmt=NULL;
+	}
 	//query.ptr=(char *)malloc(length);
 	//memcpy(query.ptr,stmt,length);
 }
@@ -445,6 +502,50 @@ void MySQL_Connection::real_query_cont(short event) {
 	async_exit_status = mysql_real_query_cont(&interr ,mysql , mysql_status(event, true));
 }
 
+void MySQL_Connection::stmt_prepare_start() {
+	PROXY_TRACE();
+	query.stmt=mysql_stmt_init(mysql);
+	my_bool my_arg=true;
+	mysql_stmt_attr_set(query.stmt, STMT_ATTR_UPDATE_MAX_LENGTH, &my_arg);
+	async_exit_status = mysql_stmt_prepare_start(&interr , query.stmt, query.ptr, query.length);
+}
+
+void MySQL_Connection::stmt_prepare_cont(short event) {
+	proxy_debug(PROXY_DEBUG_MYSQL_PROTOCOL, 6,"event=%d\n", event);
+	async_exit_status = mysql_stmt_prepare_cont(&interr , query.stmt , mysql_status(event, true));
+}
+
+void MySQL_Connection::stmt_execute_start() {
+	PROXY_TRACE();
+	int _rc=0;
+	assert(query.stmt->mysql); // if we reached here, we hit bug #740
+	_rc=mysql_stmt_bind_param(query.stmt, query.stmt_meta->binds); // FIXME : add error handling
+	if (_rc) {
+		proxy_error("mysql_stmt_bind_param() failed: %s", mysql_stmt_error(query.stmt));
+	}
+	//proxy_info("Calling mysql_stmt_execute_start, current state: %d\n", query.stmt->state);
+	async_exit_status = mysql_stmt_execute_start(&interr , query.stmt);
+	//fprintf(stderr,"Current state: %d\n", query.stmt->state);
+}
+
+void MySQL_Connection::stmt_execute_cont(short event) {
+	proxy_debug(PROXY_DEBUG_MYSQL_PROTOCOL, 6,"event=%d\n", event);
+	//proxy_info("Calling mysql_stmt_execute_cont, current state: %d\n", query.stmt->state);
+	async_exit_status = mysql_stmt_execute_cont(&interr , query.stmt , mysql_status(event, true));
+	//proxy_info("mysql_stmt_execute_cont , ret=%d\n", async_exit_status);
+	//fprintf(stderr,"Current state: %d\n", query.stmt->state);
+}
+
+void MySQL_Connection::stmt_execute_store_result_start() {
+	PROXY_TRACE();
+	async_exit_status = mysql_stmt_store_result_start(&interr, query.stmt);
+}
+
+void MySQL_Connection::stmt_execute_store_result_cont(short event) {
+	proxy_debug(PROXY_DEBUG_MYSQL_PROTOCOL, 6,"event=%d\n", event);
+	async_exit_status = mysql_stmt_store_result_cont(&interr , query.stmt , mysql_status(event, true));
+}
+
 void MySQL_Connection::store_result_start() {
 	PROXY_TRACE();
 	async_exit_status = mysql_store_result_start(&mysql_result, mysql);
@@ -455,6 +556,10 @@ void MySQL_Connection::store_result_cont(short event) {
 	async_exit_status = mysql_store_result_cont(&mysql_result , mysql , mysql_status(event, true));
 }
 
+void MySQL_Connection::set_is_client() {
+	local_stmts->set_is_client(myds->sess);
+}
+
 #define NEXT_IMMEDIATE(new_st) do { async_state_machine = new_st; goto handler_again; } while (0)
 
 MDB_ASYNC_ST MySQL_Connection::handler(short event) {
@@ -527,7 +632,8 @@ handler_again:
 			parent->connect_error(mysql_errno(mysql));
 			break;
 		case ASYNC_CONNECT_TIMEOUT:
-			proxy_error("Connect timeout on %s:%d : %llu - %llu = %llu\n",  parent->address, parent->port, myds->sess->thread->curtime , myds->wait_until, myds->sess->thread->curtime - myds->wait_until);
+			//proxy_error("Connect timeout on %s:%d : %llu - %llu = %llu\n",  parent->address, parent->port, myds->sess->thread->curtime , myds->wait_until, myds->sess->thread->curtime - myds->wait_until);
+			proxy_error("Connect timeout on %s:%d : exceeded by %lluus\n", parent->address, parent->port, myds->sess->thread->curtime - myds->wait_until);
 			parent->connect_error(mysql_errno(mysql));
 			break;
 		case ASYNC_CHANGE_USER_START:
@@ -623,6 +729,131 @@ handler_again:
 			}
 			break;
 
+		case ASYNC_STMT_PREPARE_START:
+			stmt_prepare_start();
+			__sync_fetch_and_add(&parent->queries_sent,1);
+			__sync_fetch_and_add(&parent->bytes_sent,query.length);
+			myds->sess->thread->status_variables.queries_backends_bytes_sent+=query.length;
+			if (async_exit_status) {
+				next_event(ASYNC_STMT_PREPARE_CONT);
+			} else {
+				NEXT_IMMEDIATE(ASYNC_STMT_PREPARE_END);
+			}
+			break;
+		case ASYNC_STMT_PREPARE_CONT:
+			stmt_prepare_cont(event);
+			if (async_exit_status) {
+				next_event(ASYNC_STMT_PREPARE_CONT);
+			} else {
+				NEXT_IMMEDIATE(ASYNC_STMT_PREPARE_END);
+			}
+			break;
+
+		case ASYNC_STMT_PREPARE_END:
+			if (interr) {
+				NEXT_IMMEDIATE(ASYNC_STMT_PREPARE_FAILED);
+			} else {
+				NEXT_IMMEDIATE(ASYNC_STMT_PREPARE_SUCCESSFUL);
+			}
+			break;
+		case ASYNC_STMT_PREPARE_SUCCESSFUL:
+			break;
+		case ASYNC_STMT_PREPARE_FAILED:
+			break;
+
+		case ASYNC_STMT_EXECUTE_START:
+			stmt_execute_start();
+			__sync_fetch_and_add(&parent->queries_sent,1);
+			__sync_fetch_and_add(&parent->bytes_sent,query.stmt_meta->size);
+			myds->sess->thread->status_variables.queries_backends_bytes_sent+=query.stmt_meta->size;
+//			__sync_fetch_and_add(&parent->bytes_sent,query.length);
+//			myds->sess->thread->status_variables.queries_backends_bytes_sent+=query.length;
+			if (async_exit_status) {
+				next_event(ASYNC_STMT_EXECUTE_CONT);
+			} else {
+				NEXT_IMMEDIATE(ASYNC_STMT_EXECUTE_STORE_RESULT_START);
+			}
+			break;
+		case ASYNC_STMT_EXECUTE_CONT:
+			stmt_execute_cont(event);
+			if (async_exit_status) {
+				next_event(ASYNC_STMT_EXECUTE_CONT);
+			} else {
+				NEXT_IMMEDIATE(ASYNC_STMT_EXECUTE_STORE_RESULT_START);
+			}
+			break;
+
+		case ASYNC_STMT_EXECUTE_STORE_RESULT_START:
+			if (mysql_stmt_errno(query.stmt)) {
+				NEXT_IMMEDIATE(ASYNC_STMT_EXECUTE_END);
+			}
+			{
+				query.stmt_result=mysql_stmt_result_metadata(query.stmt);
+				if (query.stmt_result==NULL) {
+					NEXT_IMMEDIATE(ASYNC_STMT_EXECUTE_END);
+				}
+			}
+			stmt_execute_store_result_start();
+			if (async_exit_status) {
+				next_event(ASYNC_STMT_EXECUTE_STORE_RESULT_CONT);
+			} else {
+				NEXT_IMMEDIATE(ASYNC_STMT_EXECUTE_END);
+			}
+			break;
+		case ASYNC_STMT_EXECUTE_STORE_RESULT_CONT:
+			stmt_execute_store_result_cont(event);
+			if (async_exit_status) {
+				next_event(ASYNC_STMT_EXECUTE_STORE_RESULT_CONT);
+			} else {
+				NEXT_IMMEDIATE(ASYNC_STMT_EXECUTE_END);
+			}
+			break;
+		case ASYNC_STMT_EXECUTE_END:
+			{
+				if (query.stmt_result) {
+					unsigned long long total_size=0;
+					MYSQL_ROWS *r=query.stmt->result.data;
+					if (r) {
+						total_size+=r->length;
+						if (r->length > 0xFFFFFF) {
+							total_size+=(r->length / 0xFFFFFF) * sizeof(mysql_hdr);
+						}
+						total_size+=sizeof(mysql_hdr);
+						while(r->next) {
+							r=r->next;
+							total_size+=r->length;
+							if (r->length > 0xFFFFFF) {
+								total_size+=(r->length / 0xFFFFFF) * sizeof(mysql_hdr);
+							}
+							total_size+=sizeof(mysql_hdr);
+						}
+					}
+					__sync_fetch_and_add(&parent->bytes_recv,total_size);
+					myds->sess->thread->status_variables.queries_backends_bytes_recv+=total_size;
+				}
+/*
+				int row_count= 0;
+				fprintf(stdout, "Fetching results ...\n");
+				while (!mysql_stmt_fetch(query.stmt))
+				{
+					row_count++;
+					fprintf(stdout, "  row %d\n", row_count);
+				}
+*/
+			}
+/*
+			if (interr) {
+				NEXT_IMMEDIATE(ASYNC_STMT_EXECUTE_FAILED);
+			} else {
+				NEXT_IMMEDIATE(ASYNC_STMT_EXECUTE_SUCCESSFUL);
+			}
+*/
+			break;
+//		case ASYNC_STMT_EXECUTE_SUCCESSFUL:
+//			break;
+//		case ASYNC_STMT_EXECUTE_FAILED:
+//			break;
+
 		case ASYNC_NEXT_RESULT_START:
 			async_exit_status = mysql_next_result_start(&interr, mysql);
 			if (async_exit_status) {
@@ -710,6 +941,16 @@ handler_again:
 						NEXT_IMMEDIATE(ASYNC_USE_RESULT_CONT); // we continue looping
 					}
 				} else {
+					if (mysql) {
+						int _myerrno=mysql_errno(mysql);
+						if (_myerrno) {
+							if (myds) {
+								MyRS->add_err(myds);
+								NEXT_IMMEDIATE(ASYNC_QUERY_END);
+							}
+						}
+					}
+					// we reach here if there was no error
 					MyRS->add_eof();
 					NEXT_IMMEDIATE(ASYNC_QUERY_END);
 				}
@@ -749,6 +990,10 @@ handler_again:
 			}
 			break;
 		case ASYNC_SET_AUTOCOMMIT_SUCCESSFUL:
+			options.last_set_autocommit = ( options.autocommit ? 1 : 0 ) ; // we successfully set autocommit
+			if ((mysql->server_status & SERVER_STATUS_AUTOCOMMIT) && options.autocommit==false) {
+				proxy_warning("It seems we are hitting bug http://bugs.mysql.com/bug.php?id=66884\n");
+			}
 			break;
 		case ASYNC_SET_AUTOCOMMIT_FAILED:
 			fprintf(stderr,"%s\n",mysql_error(mysql));
@@ -898,14 +1143,17 @@ int MySQL_Connection::async_connect(short event) {
 // 0 when the query is completed
 // 1 when the query is not completed
 // the calling function should check mysql error in mysql struct
-int MySQL_Connection::async_query(short event, char *stmt, unsigned long length) {
+int MySQL_Connection::async_query(short event, char *stmt, unsigned long length, MYSQL_STMT **_stmt, stmt_execute_metadata_t *stmt_meta) {
 	PROXY_TRACE();
 	assert(mysql);
 	assert(ret_mysql);
+	server_status=parent->status; // we copy it here to avoid race condition. The caller will see this
 	if (
-		(parent->status==MYSQL_SERVER_STATUS_OFFLINE_HARD) // the server is OFFLINE as specific by the user
+		(server_status==MYSQL_SERVER_STATUS_OFFLINE_HARD) // the server is OFFLINE as specific by the user
 		||
-		(parent->status==MYSQL_SERVER_STATUS_SHUNNED && parent->shunned_automatic==true && parent->shunned_and_kill_all_connections==true) // the server is SHUNNED due to a serious issue
+		(server_status==MYSQL_SERVER_STATUS_SHUNNED && parent->shunned_automatic==true && parent->shunned_and_kill_all_connections==true) // the server is SHUNNED due to a serious issue
+		||
+		(server_status==MYSQL_SERVER_STATUS_SHUNNED_REPLICATION_LAG) // slave is lagging! see #774
 	) {
 		return -1;
 	}
@@ -915,8 +1163,20 @@ int MySQL_Connection::async_query(short event, char *stmt, unsigned long length)
 			return 0;
 			break;
 		case ASYNC_IDLE:
-			set_query(stmt,length);
+			if (stmt_meta==NULL)
+				set_query(stmt,length);
 			async_state_machine=ASYNC_QUERY_START;
+			if (_stmt) {
+				query.stmt=*_stmt;
+				if (stmt_meta==NULL) {
+					async_state_machine=ASYNC_STMT_PREPARE_START;
+				} else {
+					if (query.stmt_meta==NULL) {
+						query.stmt_meta=stmt_meta;
+					}
+					async_state_machine=ASYNC_STMT_EXECUTE_START;
+				}
+			}
 		default:
 			handler(event);
 			break;
@@ -929,6 +1189,26 @@ int MySQL_Connection::async_query(short event, char *stmt, unsigned long length)
 			return 0;
 		}
 	}
+	if (async_state_machine==ASYNC_STMT_EXECUTE_END) {
+		query.stmt_meta=NULL;
+		async_state_machine=ASYNC_QUERY_END;
+		if (mysql_stmt_errno(query.stmt)) {
+			return -1;
+		} else {
+			return 0;
+		}
+	}
+	if (async_state_machine==ASYNC_STMT_PREPARE_SUCCESSFUL || async_state_machine==ASYNC_STMT_PREPARE_FAILED) {
+		query.stmt_meta=NULL;
+		if (async_state_machine==ASYNC_STMT_PREPARE_FAILED) {
+			//mysql_stmt_close(query.stmt);
+			//query.stmt=NULL;
+			return -1;
+		} else {
+			*_stmt=query.stmt;
+			return 0;
+		}
+	}
 	if (async_state_machine==ASYNC_NEXT_RESULT_START) {
 		// if we reached this point it measn we are processing a multi-statement
 		// and we need to exit to give control to MySQL_Session
@@ -1141,6 +1421,16 @@ void MySQL_Connection::async_free_result() {
 		query.ptr=NULL;
 		query.length=0;
 	}
+	if (query.stmt_result) {
+		mysql_free_result(query.stmt_result);
+		query.stmt_result=NULL;
+	}
+	if (query.stmt) {
+		if (query.stmt->mysql) {
+			mysql_stmt_free_result(query.stmt);
+		}
+		query.stmt=NULL;
+	}
 	if (mysql_result) {
 		mysql_free_result(mysql_result);
 		mysql_result=NULL;
@@ -1156,6 +1446,9 @@ bool MySQL_Connection::IsActiveTransaction() {
 	bool ret=false;
 	if (mysql) {
 		ret = (mysql->server_status & SERVER_STATUS_IN_TRANS);
+		if (ret == false && (mysql)->net.last_errno) {
+			ret = true;
+		}
 	}
 	return ret;
 }
@@ -1164,6 +1457,14 @@ bool MySQL_Connection::IsAutoCommit() {
 	bool ret=false;
 	if (mysql) {
 		ret = (mysql->server_status & SERVER_STATUS_AUTOCOMMIT);
+		if (ret) {
+			if (options.last_set_autocommit==0) {
+				// it seems we hit bug http://bugs.mysql.com/bug.php?id=66884
+				// we last sent SET AUTOCOMMIT = 0 , but the server says it is 1
+				// we assume that what we sent last is correct .  #873
+				ret = false;
+			}
+		}
 	}
 	return ret;
 }
@@ -1172,7 +1473,7 @@ bool MySQL_Connection::MultiplexDisabled() {
 // status_flags stores information about the status of the connection
 // can be used to determine if multiplexing can be enabled or not
 	bool ret=false;
-	if (status_flags & (STATUS_MYSQL_CONNECTION_TRANSACTION|STATUS_MYSQL_CONNECTION_USER_VARIABLE|STATUS_MYSQL_CONNECTION_PREPARED_STATEMENT|STATUS_MYSQL_CONNECTION_LOCK_TABLES|STATUS_MYSQL_CONNECTION_TEMPORARY_TABLE|STATUS_MYSQL_CONNECTION_GET_LOCK) ) {
+	if (status_flags & (STATUS_MYSQL_CONNECTION_TRANSACTION|STATUS_MYSQL_CONNECTION_USER_VARIABLE|STATUS_MYSQL_CONNECTION_PREPARED_STATEMENT|STATUS_MYSQL_CONNECTION_LOCK_TABLES|STATUS_MYSQL_CONNECTION_TEMPORARY_TABLE|STATUS_MYSQL_CONNECTION_GET_LOCK|STATUS_MYSQL_CONNECTION_FOUND_ROWS|STATUS_MYSQL_CONNECTION_SQL_LOG_BIN0) ) {
 		ret=true;
 	}
 	return ret;
@@ -1191,9 +1492,15 @@ void MySQL_Connection::ProcessQueryAndSetStatusFlags(char *query_digest_text) {
 				set_status_user_variable(true);
 			}
 		}
-		// For issue #555 , multiplexing is disabled if --safe-updates is used
-		if (strcasecmp(query_digest_text,"SET SQL_SAFE_UPDATES=?,SQL_SELECT_LIMIT=?,MAX_JOIN_SIZE=?")==0) {
+		if (strncasecmp(query_digest_text,"SET ",4)==0) {
+			// For issue #555 , multiplexing is disabled if --safe-updates is used
+			if (strcasecmp(query_digest_text,"SET SQL_SAFE_UPDATES=?,SQL_SELECT_LIMIT=?,MAX_JOIN_SIZE=?")==0) {
 				set_status_user_variable(true);
+			} else if (strcasecmp(query_digest_text,"SET FOREIGN_KEY_CHECKS=?")==0) { // see #835
+				set_status_user_variable(true);
+			} else if (strcasecmp(query_digest_text,"SET UNIQUE_CHECKS=?")==0) { // see #835
+				set_status_user_variable(true);
+			}
 		}
 	}
 	if (get_status_prepared_statement()==false) { // we search if prepared was already executed
@@ -1226,6 +1533,11 @@ void MySQL_Connection::ProcessQueryAndSetStatusFlags(char *query_digest_text) {
 			set_status_get_lock(true);
 		}
 	}
+	if (get_status_found_rows()==false) { // we search for SQL_CALC_FOUND_ROWS if not already set
+		if (strcasestr(query_digest_text,"SQL_CALC_FOUND_ROWS")) {
+			set_status_found_rows(true);
+		}
+	}
 }
 
 void MySQL_Connection::optimize() {
