@@ -128,8 +128,11 @@ using namespace std;
 static volatile int load_;
 
 //__thread l_sfp *__thr_sfp=NULL;
-
+//#ifdef DEBUG
+//const char *malloc_conf = "xmalloc:true,lg_tcache_max:16,purge:decay,junk:true,tcache:false";
+//#else
 const char *malloc_conf = "xmalloc:true,lg_tcache_max:16,purge:decay";
+//#endif /* DEBUG */
 //const char *malloc_conf = "prof_leak:true,lg_prof_sample:0,prof_final:true,xmalloc:true,lg_tcache_max:16";
 
 int listen_fd;
@@ -142,6 +145,8 @@ Query_Processor *GloQPro;
 ProxySQL_Admin *GloAdmin;
 MySQL_Threads_Handler *GloMTH;
 
+MySQL_STMT_Manager *GloMyStmt;
+
 MySQL_Monitor *GloMyMon;
 std::thread *MyMon_thread;
 
@@ -167,6 +172,26 @@ void * mysql_worker_thread_func(void *arg) {
 	return NULL;
 }
 
+void * mysql_worker_thread_func_idles(void *arg) {
+
+//	__thr_sfp=l_mem_init();
+	proxysql_mysql_thread_t *mysql_thread=(proxysql_mysql_thread_t *)arg;
+	MySQL_Thread *worker = new MySQL_Thread();
+	mysql_thread->worker=worker;
+	worker->epoll_thread=true;
+	worker->init();
+//	worker->poll_listener_add(listen_fd);
+//	worker->poll_listener_add(socket_fd);
+	__sync_fetch_and_sub(&load_,1);
+	do { usleep(50); } while (load_);
+
+	worker->run();
+	//delete worker;
+	delete worker;
+//	l_mem_destroy(__thr_sfp);
+	return NULL;
+}
+
 void * mysql_shared_query_cache_funct(void *arg) {
 	GloQC->purgeHash_thread(NULL);
 	return NULL;
@@ -179,6 +204,8 @@ void ProxySQL_Main_process_global_variables(int argc, const char **argv) {
 	// alwasy try to open a config file
 	if (GloVars.confFile->OpenFile(GloVars.config_file) == true) {
 		GloVars.configfile_open=true;
+	} else {
+		proxy_warning("Unable to open config file %s\n", GloVars.config_file); // issue #705
 	}
 	char *t=getcwd(NULL, 512);
 	if (GloVars.__cmd_proxysql_datadir==NULL) {
@@ -236,9 +263,11 @@ void ProxySQL_Main_init_main_modules() {
 	GloMyAuth=NULL;
 	GloMyMon=NULL;
 	GloMyLogger=NULL;
+	GloMyStmt=NULL;
 	MyHGM=new MySQL_HostGroups_Manager();
 	GloMTH=new MySQL_Threads_Handler();
 	GloMyLogger = new MySQL_Logger();
+	GloMyStmt=new MySQL_STMT_Manager();
 }
 
 
@@ -264,9 +293,10 @@ void ProxySQL_Main_init_Query_module() {
 void ProxySQL_Main_init_MySQL_Threads_Handler_module() {
 	unsigned int i;
 	GloMTH->init();
-	load_ = GloMTH->num_threads + 1;
+	load_ = GloMTH->num_threads * 2 + 1;
 	for (i=0; i<GloMTH->num_threads; i++) {
-		GloMTH->create_thread(i,mysql_worker_thread_func);
+		GloMTH->create_thread(i,mysql_worker_thread_func, false);
+		GloMTH->create_thread(i,mysql_worker_thread_func_idles, true);
 	}
 }
 
@@ -387,6 +417,10 @@ void ProxySQL_Main_shutdown_all_modules() {
 		std::cerr << "GloHGM shutdown in ";
 #endif
 	}
+	if (GloMyStmt) {
+		delete GloMyStmt;
+		GloMyStmt=NULL;
+	}
 }
 
 void ProxySQL_Main_init() {
@@ -693,6 +727,8 @@ gotofork:
 
 
 
+	} else {
+		GloAdmin->flush_error_log();
 	}
 
 __start_label:
